unit FileList;

interface

type
  TFileList = class(TObject)
  public
    Cur: Integer;
    Count: Integer;
    Items: array of WideString;
    ByteItems: array of string; // for archive plugin
    constructor Create;
    procedure Add(const Path: WideString); overload;
    procedure Add(const Path: string); overload;
    procedure AddDir(const Path: WideString);
    procedure Clear;
    procedure Delete(Index: Integer);
    procedure Sort;
  end;

implementation

uses ShlwapiSub, NLibW, NSysW;

constructor TFileList.Create;
begin
  Clear;
end;

procedure TFileList.Add(const Path: WideString);
begin
  if Count = Length(Items) then
  begin
    SetLength(Items, Count + Count div 4);
    SetLength(ByteItems, Count + Count div 4);
  end;
  Items[Count] := Path;
  ByteItems[Count] := Path;
  Inc(Count);
end;

procedure TFileList.Add(const Path: string);
begin
  if Count = Length(Items) then
  begin
    SetLength(Items, Count + Count div 4);
    SetLength(ByteItems, Count + Count div 4);
  end;
  Items[Count] := Path;
  ByteItems[Count] := Path;
  Inc(Count);
end;

procedure TFileList.AddDir(const Path: WideString);
var
  dir: TDir;
begin
  with TListDir.Create(Path) do
    try
      while Each(dir) do
        if not (dmDirectory in dir.Mode) then
          Add(JoinPath([Path, dir.Name]));
    finally
      Free;
    end;
end;

procedure TFileList.Clear;
begin
  Count := 0;
  SetLength(Items, 32);
  SetLength(ByteItems, 32);
end;

procedure TFileList.Delete(Index: Integer);
var
  i: Integer;
begin
  if (Index < 0) or (Index >= Count) then
    Exit;
  Dec(Count);
  for i := Index to Count - 1 do
  begin
    Items[i] := Items[i+1];
    ByteItems[i] := ByteItems[i+1];
  end;
end;

function StrCmpLogicalWrapper(const A: string; const B: string): Integer;
begin
  Result := StrCmpLogicalW(PWideChar(WideString(A)), PWideChar(WideString(B)));
end;

function StrCmpLogicalWrapperW(const A: WideString; const B: WideString): Integer;
begin
  Result := StrCmpLogicalW(PWideChar(A), PWideChar(B));
end;

procedure TFileList.Sort;
var
  compare: function (const A: string; const B: string): Integer;
  comparew: function (const A: WideString; const B: WideString): Integer;
begin
  if Count = 0 then Exit;

  if Assigned(StrCmpLogicalW) then
  begin
    compare := StrCmpLogicalWrapper;
    comparew := StrCmpLogicalWrapperW;
  end
  else begin
    compare := StrColl;
    comparew := StrCollW;
  end;
  SetLength(Items, Count);
  NLibW.Sort(Items, comparew);
  SetLength(ByteItems, Count);
  NLibW.Sort(ByteItems, compare);
end;

end.
